/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2008 Bull S.A.S.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: CMIServerConfigurator.java 2038 2008-09-05 16:58:04Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.controller.server.component;

import java.io.File;
import java.net.URL;

import org.ow2.cmi.config.CMIConfig;
import org.ow2.cmi.controller.server.ServerClusterViewManager;
import org.ow2.cmi.controller.server.ServerClusterViewManagerException;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.url.URLUtils;
import org.ow2.util.url.URLUtilsException;
import org.ow2.util.xmlconfig.XMLConfiguration;
import org.ow2.util.xmlconfig.XMLConfigurationException;
import org.ow2.util.xmlconfig.properties.SystemPropertyResolver;


/**
 * Allows to configure a CMI server with an XML configuration file.
 * @author Loris Bouzonnet (greatly inspired from org.ow2.easybeans.server.EmbeddedConfigurator by Florent Benoit)
 */
public final class CMIServerConfigurator {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(CMIServerConfigurator.class);


    /**
     * Utility class, no public constructor.
     */
    private CMIServerConfigurator() {
    }

    /**
     * Configure the given CMI server with the xml configuration file.
     * @param serverClusterViewManager the CMI server to configure.
     * @throws ServerClusterViewManagerException if the configuration fails.
     */
    public static void configure(final ServerClusterViewManager serverClusterViewManager)
    throws ServerClusterViewManagerException {
        String confDir = CMIConfig.getConfDir();
        String confFilename = CMIConfig.getServerConfFilename();
        URL url = null;
        if(confDir != null) {
            File file = new File(confDir, confFilename);
            if(file.exists()) {
                try {
                    url = URLUtils.fileToURL2(file);
                } catch (URLUtilsException e) {
                    logger.warn("Cannot get the url for the file: {0}", file, e);
                }
            } else {
                logger.warn("File not found: {0}", file);
            }
        }
        if(url == null) {
            url = ClassLoader.getSystemClassLoader().getResource(confFilename);
            if(url == null) {
                url = Thread.currentThread().getContextClassLoader().getResource(confFilename);
                if(url == null) {
                    url = serverClusterViewManager.getClass().getResource(confFilename);
                }
                if(url == null) {
                    logger.info("Unable to find cmi-server-config.xml: using the default configuration.");

                }
            }
        }
        long tStart = System.currentTimeMillis();
        logger.debug("Starting configuration of CMI server");
        XMLConfiguration xmlConfiguration = new XMLConfiguration(url, "cmi-server-mapping.xml");

        // Resolve properties as system properties
        xmlConfiguration.setPropertyResolver(new SystemPropertyResolver());

        try {
            xmlConfiguration.configure(serverClusterViewManager);
        } catch (XMLConfigurationException e) {
            logger.error("Cannot configure the CMI server", e);
            throw new ServerClusterViewManagerException("Cannot configure the CMI server", e);
        }
        if (logger.isInfoEnabled()) {
            logger.debug("Configuration done in : " + (System.currentTimeMillis() - tStart) + " ms");
        }
    }

}
